package Event::XML::Sync;

use warnings;
use strict;
use Carp;

use Event qw/loop unloop/;
use IPC::Open2;
use XML::Twig;
use IO::Pipe;

use constant (BUFSIZE => 32768);
use UNIVERSAL qw(isa can);

# --- module variables --------------------------------------------------------

use vars qw($VERSION);
$VERSION = '1.0';

# --- module interface --------------------------------------------------------

sub new {
  my $class = shift;
  my $opts = shift;

  my $self = bless({}, $class);

  my %opts = !defined($opts) ? () : ref($opts) ? %$opts : ($opts, @_);
  %$self = (%$self, %opts);

  exists $opts{cmd} || croak "cmd option expected";

  # defaults
  $self->{callbacks}{done} ||= \&done;
  $self->{data} ||= \*STDIN;

  # Start a new twig with reentrency on $sync
  $self->{twig} = XML::Twig->new(@{$self->{twig}});
  $self->{twig}{sync} = $self;

  # Start $cmd
  my ($sender,$receiver);
  $self->{pid} = open2($sender,$receiver,$self->{cmd}) 
    || croak "Could not start $self->{cmd}";

  # Set IO Event Watcher on $receiver
  $self->{receive_watcher} =
    Event->io(desc=> "pipeline reader to sync",
	      fd=>$sender,
	      poll=> 'r',
	      cb=>\&sync,
	      data=>$self);
  $self->debug("Set IO Watcher");

  # Set Idle Event Watcher on reading $handle
  $self->{data_watcher} =
    Event->io(desc => "date reader to feed XML parser",
	      fd=>$self->{data},
	      poll=>'r',
	      cb=>\&get_data,
	      data=>$self);
  $self->debug("Set Input Data Watcher");

  $self->{receiver}=$receiver;
  $self->{sender}=$sender;
  $self->{done} = 0;
  $self->{fifo} = [];

  my @parse_args=();
  @parse_args = @{$opts{parser}} if (exists $opts{parser});

  # Start parsing XML
  $self->{twig}->parse_start(@parse_args);
  $self->debug("Starting Twig Parser");

##  return $self;

  loop;

}

sub store {
  my $self = shift;
  my $element = shift;
  unshift(@{$self->{fifo}},$element);
}

sub delayed_run {
  my $self = shift;
  my $element = shift;
  if (@{$self->{fifo}}) {
    unshift(@{$self->{fifo}},$element);
  } else {
    $element->run;
  }
}

sub get {
  my $self = shift;
  return pop(@{$self->{fifo}});
}

sub write {
  my $self = shift;
  $self->{receiver}->print(@_);
}

sub flush {
  my $self = shift;
  my $element = shift;
  $self->{twig}->flush_up_to($element);
}

sub process {
  my $self = shift;
  my $element = shift;
  $self->store([$element]);
  $self->write(@_);
}

sub debug {
  my $self = shift;
  return unless $self->{debug};
  print STDERR "*** @_\n";
}

sub done {
  my $self = shift;
  $self->{twig}->flush();
}

# --- event callbacks --------------------------------------------------------

sub get_data {
  my $self = shift;		# an event
  my $handle = $self->w->fd;
  my $sync = $self->w->data;
  my $twig = $sync->{twig};
  if (eof($handle)) {
    $sync->debug("Done data");
    # All data read => cancel data watcher
    $twig->parser && $twig->parser->parse_done;	# close XML parser
    close($handle);		# close data handle
    close($sync->{receiver});	# close XML parser input
    $sync->{done} = 1;	        # set flag
    $self->w->cancel;		# remove watcher
    if (!@{$sync->{fifo}}) {	# exit if nothing to be done
      $sync->{callbacks}{done}->($sync);
      close($sync->{sender});	# close pipeline output (and get buffered data)
      unloop;
    }
  } else {
    # Parse one more line
    my $line = <$handle>;
    $sync->debug("Getting data: $line");
    $twig->parser->parse_more($line);
  }
}

sub sync {
  my $self=shift;		# an event
  my $sync = $self->w->data;
  my $handle = $self->w->fd;
  $sync->debug("Synchronizing");
  return  if eof($handle);
  my $input = <$handle>;
  my $data = $sync->{fifo}->[-1];
  push(@$data,$input);
  my $obj = $data->[0];
  my @inputs = @$data;
  shift(@inputs);
  # by default, sync after every returned line
  $sync->debug("Check sync @inputs");
  return unless (!can($obj,'check_sync') || $obj->check_sync(@inputs));
  pop(@{$sync->{fifo}});
  $obj->sync($sync,@inputs);
  # Delayed actions may now be ran
  while(@{$sync->{fifo}}) {
    my $obj = $sync->{fifo}->[-1];
    last unless (can($obj,'run'));
    $obj->run;
    pop(@{$sync->{fifo}});
  }
  if ($sync->{done} && !@{$sync->{fifo}}) {
    $sync->{callbacks}{done}->($sync);
    $self->w->cancel;
    close($sync->{sender});	# close pipeline output (and get buffered data)
    unloop;
  }
}

1;

__END__

=head1 NAME

Event::XML::Sync - Synchronizing XML data using module Event

=head1 SYNOPSIS

use Event::XML::Sync;

Event::XML::Sync->new( cmd => "mytagger",
                       twig =>  [ 
                        twig_handlers => { token => \&token_process }] 
                     );

sub token_process {
  my ($t,$token) = @_;
  my $self = $t->{sync};
  my $form = $token->att('form');
  $self->process($token,"$form\n");
}

sub XML::Twig::Elt::sync {
  my $token = shift;
  my $sync = shift;
  my $input = shift;
  my @info = split(/\t+/,$input);
  $token->set_att('tag',$info[1]);
  $sync->flush($token);
}

=head1 DESCRIPTION

This module may be used to receive an XML stream analyzed by
XML::Twig, using XML::Twig handlers to send bits of information to a
child process, get some information back from the child process and
synchronize this information within the original stream.

=head1 METHODS

The module provides an object-oriented interface as follows:

=head2 new [hash-ref]

Used to create and run an event loop, start the child process and
start parsing the input stream. The function may be passed either a
hash or a reference to a hash. The keys below are meaningful to the
component, all others are passed to the provided callbacks.

=item cmd [string]

The child command.

=item data [Stream]

The input data (by default <>).

=item callbacks [hash-ref]

This hash reference contains mappings for the events the component
will generate. Callers can set these values to callbacks (code
references). For instance, a callback for event 'done' may be set,
triggered when the child process is closed (the default done callback
flush the remaining twig to output, printing the remaining XML parts).

=item twig [hash-ref]

This hash reference contains mapping to be sent to the Twig component
(see XML::Twig).

=item parser

This list reference contains mapping to be sent to the XML parser (see
XML::Parser).

=cut

=head1 process  [closure] [in] 

This method is used to send information B<in> to the child process.
Upon reception of information B<out> returned by the child process,
method B<sync> will be called on B<closure> with argument B<sync> (the
synchronization structure) and B<out>.

=head1 delayed_run [closure]

The method is used to delay the execution of an action until
completion of pending synchronized actions (started with method
process). When there is no pending synchronized action, method B<run>
is called on B<closure>.

=head1 AUTHOR

Eric de la Clergerie <Eric.De_La_Clergerie@inria.fr>

=head1 DATE

$Date: 2004/02/19 14:10:35 $

=head1 VERSION

$Revision: 1.2 $

=head1 LICENSE AND COPYRIGHT

=cut
